﻿// Windows (VisualStudio)
//  cl /LD /Feruler.wfm ruler.c opengl32.lib user32.lib gdi32.lib
//
// Linux (GCC)
//  gcc -O -shared -fPIC -o ruler.lfm ruler.c

///////////////////////////////////////////////////////////////////////////////
// ヘッダーファイル
#include <stdio.h>
#include <wchar.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#ifdef _WIN32
#include <Windows.h>
#else
#include <X11/Xlib.h>
#include <GL/glx.h>
#endif
#include <GL/gl.h>

///////////////////////////////////////////////////////////////////////////////
// Windows用エクスポート宣言
#ifdef _WIN32
#ifndef _WIN64
#error 64bitバイナリーとしてビルドしてください
#endif
#define EXPORT __declspec(dllexport)
#else
#define EXPORT
#endif
///////////////////////////////////////////////////////////////////////////////
// 定数(タイプ)
#define	TYPE_8      0x0001
#define	TYPE_16     0x0002
#define	TYPE_32     0x0004
#define	TYPE_24     0x0008
#define	TYPE_48     0x0010
#define	TYPE_96     0x0020
#define	TYPE_192    0x0040
#define	TYPE_UI     0x0100

///////////////////////////////////////////////////////////////////////////////
// 定数
#define	OK          1
#define	NG          0
#define	ON          1
#define	OFF         0
#define	DIALOG_W    200
#define	DIALOG_H    70
#define CAPTION_JP  "定規"
#define CAPTION_EN  "Ruler"
#define	FORMAT_JP   "距離 %.2f%ls"
#define	FORMAT_EN   "distance %.2%ls"
#ifdef _WIN32
#define CLASS_NAME  "ruler_dialog"
#endif

///////////////////////////////////////////////////////////////////////////////
// 構造体
typedef struct _XWND {
#ifdef _WIN32
	HWND window;
#else
	Display *display;
	Window window;
#endif
} XWND;
typedef struct _EFM {
	// インスタンス毎に保持するデータ
	void *instance;
	XWND dialog;
	char msg[256];
	int sx, sy, ex, ey, mouse_on;
} EFM;

///////////////////////////////////////////////////////////////////////////////
// グローバル変数
const static char *g_caption = CAPTION_JP;
const static char *g_format = FORMAT_JP;
#ifndef _WIN32
static Atom atom_t;
static Atom atom_d;
#endif

///////////////////////////////////////////////////////////////////////////////
// 外部関数
static void   (*upcall_redraw_view)(void) = NULL;                           // UI操作でoverlay_viewの変わった時に再描画要求する
static void   (*upcall_close_ui)(void* instance) = NULL;                    // UIのクローズボタンでUIを閉じた事を通知する
static void   (*upcall_set_videoframe)(size_t frame) = NULL;                // 動画のフレームを変更する
static double (*upcall_getscale)(int* axis_y, const wchar_t **unit) = NULL; // 計測スケールと単位・座標系を取得 戻り値:ピクセル→実座標の係数 axis_y:1=Y軸下向き -1:Y軸上向き unit:単位文字列

///////////////////////////////////////////////////////////////////////////////
// 内部関数
#ifdef _WIN32
//=============================================================================
// 
//=============================================================================
LRESULT CALLBACK WndProc(HWND hWnd, UINT mes, WPARAM wParam, LPARAM lParam)
{
	EFM *efm;
	HDC hDC;
	PAINTSTRUCT ps;

	// 関連付けられたEFMを取得
	efm = (EFM*)GetWindowLongPtr(hWnd, 0);

	// 入力メッセージ処理
	switch (mes) {
	case WM_PAINT:
		hDC = BeginPaint(hWnd, &ps);
		TextOut(hDC, 0, 0, efm->msg, strlen(efm->msg));
		EndPaint(hWnd, &ps);
		return 0;

	case WM_CLOSE:
		upcall_close_ui(efm->instance);
		return 0;

	default:
		break;
	}

	// デフォルト処理
	return DefWindowProc(hWnd, mes, wParam, lParam);
}

//=============================================================================
// 
//=============================================================================
static void window_caption(XWND xwnd, const char *caption)
{
	SetWindowText(xwnd.window, caption);
}

//=============================================================================
// 
//=============================================================================
static XWND window_open(XWND parent, const char *caption, void *ex_info)
{
	XWND xwnd;
	HINSTANCE hInstance;
	WNDCLASSEX wcex;
	DWORD wstyle;
	static int class_registed = OFF;
	RECT rect;

	// 作成
	hInstance = GetModuleHandle(NULL);
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(WNDCLASSEX);
	wcex.style = CS_OWNDC | CS_HREDRAW | CS_VREDRAW;
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.lpfnWndProc = WndProc;
	wcex.hInstance = hInstance;
	wcex.lpszClassName = CLASS_NAME;
	wcex.cbWndExtra = sizeof(LONG_PTR);
	if (!class_registed) {
		RegisterClassEx(&wcex);
		class_registed = ON;
	}
	wstyle = WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU;
	rect.left = 0;
	rect.top = 0;
	rect.right = (long)(DIALOG_W - 1);
	rect.bottom = (long)(DIALOG_H - 1);
	AdjustWindowRect(&rect, wstyle, FALSE);
	xwnd.window = CreateWindow(CLASS_NAME, caption, wstyle, CW_USEDEFAULT, 0, rect.right-rect.left+1, rect.bottom-rect.top+1, parent.window, NULL, hInstance, NULL);
	if (xwnd.window) {
		SetWindowLongPtr(xwnd.window, 0, (LONG_PTR)ex_info);
	}

	// 表示
	ShowWindow(xwnd.window, SW_SHOW);

	// 終了
	return xwnd;
}

//=============================================================================
// 
//=============================================================================
static void window_close(XWND xwnd)
{
	DestroyWindow(xwnd.window);
}

//=============================================================================
// 
//=============================================================================
static void window_paint(XWND xwnd)
{
	InvalidateRect(xwnd.window, NULL, ON);
}
#else
//=============================================================================
// 
//=============================================================================
static int event_func(XEvent xev, void *efmh)
{
	EFM *efm;
	GC gc;
	XFontSet font;
	int missing_char_count;
	char **missing_char_list;
	char *def_string;
	unsigned long fg, bg;

	// パラメータ取得
	efm = (EFM*)efmh;

	// 処理
	switch (xev.type) {
	case DestroyNotify:
		// 終了
		return NG;

	case ClientMessage:
		if ((xev.xclient.message_type == atom_t) && (xev.xclient.data.l[0] == atom_d)) {
			upcall_close_ui(efm->instance);
		}
		break;

	case Expose:
		fg = BlackPixel(xev.xany.display, 0);
		bg = WhitePixel(xev.xany.display, 0);
		gc = XCreateGC(xev.xany.display, xev.xany.window, 0, NULL);
		XSetBackground(xev.xany.display, gc, bg);
		XSetForeground(xev.xany.display, gc, bg);
		XFillRectangle(xev.xany.display, xev.xany.window, gc, 0, 0, DIALOG_W, DIALOG_H);
		font = XCreateFontSet(xev.xany.display, "-*-fixed-medium-r-normal--16-*-*-*", &missing_char_list, &missing_char_count, &def_string);
		if (font) {
			if (missing_char_list) {
				XFreeStringList(missing_char_list);
			}
			XSetForeground(xev.xany.display, gc, fg);
			XmbDrawImageString(xev.xany.display, xev.xany.window, font, gc, 0, 0+16, efm->msg, strlen(efm->msg));
			XFreeFontSet(xev.xany.display, font);
		}
		XFreeGC(xev.xany.display, gc);
		XFlush(xev.xany.display);
		break;

	default:
		break;
	}

	// 継続
	return OK;
}

//=============================================================================
// 
//=============================================================================
static void window_caption(XWND xwnd, const char *caption)
{
	XTextProperty tp;
	char *tmp_caption;

	// キャプション設定
	tmp_caption = (char*)malloc(sizeof(char) * (strlen(caption) + 1));
	if (tmp_caption) {
		strncpy(tmp_caption, caption, strlen(caption) + 1);
		XmbTextListToTextProperty(xwnd.display, &tmp_caption, 1, XCompoundTextStyle, &tp);
		XSetWMName(xwnd.display, xwnd.window, &tp);
		XFree(tp.value);
		free(tmp_caption);
	}
	else {
		XStoreName(xwnd.display, xwnd.window, "");
	}
	XFlush(xwnd.display);
}

//=============================================================================
// 
//=============================================================================
static XWND window_open(XWND parent, const char *caption, void *ex_info)
{
	XWND xwnd;
	EFM *efm;
	int nof_vi;
	XSetWindowAttributes xattr;
	unsigned long attrmask;
	XSizeHints *hint;

	// 初期化
	memset(&xwnd, 0, sizeof(XWND));
	efm = (EFM*)ex_info;

	// 作成
	if (parent.display) {
		xwnd.display = parent.display;
	}
	else {
		xwnd.display = XOpenDisplay(NULL);
	}
	if (!xwnd.display) {
		return xwnd;
	}
	memset(&xattr, 0, sizeof(xattr));
	xattr.override_redirect = 0;
	attrmask = CWOverrideRedirect;
	xwnd.window = XCreateWindow(xwnd.display, DefaultRootWindow(xwnd.display), 0, 0, DIALOG_W, DIALOG_H, 1, 24, InputOutput, CopyFromParent, attrmask, &xattr);
	if (xwnd.window) {
		XSetTransientForHint(xwnd.display, xwnd.window, parent.window);
		atom_t = XInternAtom(xwnd.display, "WM_PROTOCOLS", False);
		atom_d = XInternAtom(xwnd.display, "WM_DELETE_WINDOW", False);
		XSetWMProtocols(xwnd.display, xwnd.window, &atom_d, 1);

		hint = XAllocSizeHints();
		if (hint) {
			hint->flags = PMinSize | PMaxSize;
			hint->min_width = hint->max_width = DIALOG_W;
			hint->min_height = hint->max_height = DIALOG_H;
			XSetWMNormalHints(xwnd.display, xwnd.window, hint);
			XFree(hint);
		}

		XSelectInput(xwnd.display, xwnd.window,
                     KeyPressMask|KeyReleaseMask|
                     ButtonPressMask|ButtonReleaseMask|PointerMotionMask|
                     Button1MotionMask|Button2MotionMask|Button3MotionMask|
                     ExposureMask|StructureNotifyMask|SubstructureNotifyMask);
		XMapWindow(xwnd.display, xwnd.window);
	}
	else {
		XCloseDisplay(xwnd.display);
		xwnd.display = NULL;
		return xwnd;
	}

	// キャプション設定
	window_caption(xwnd, caption);

	// 終了
	return xwnd;
}

//=============================================================================
// 
//=============================================================================
static void window_close(XWND xwnd)
{
	// ウインドウを閉じる
	XDestroyWindow(xwnd.display, xwnd.window);
}

//=============================================================================
// 
//=============================================================================
static void window_paint(XWND xwnd)
{
	XEvent xev;

	// Exposuイベントを送信
	memset(&xev, 0, sizeof(xev));
	xev.type = Expose;
	xev.xexpose.window = xwnd.window;
	XSendEvent(xwnd.display, xwnd.window, False, ExposureMask, &xev);
	XFlush(xwnd.display);
}
#endif

//=============================================================================
// 
//=============================================================================
static int window_enable(XWND xwnd)
{
	return xwnd.window ? OK : NG;
}

//=============================================================================
// 
//=============================================================================
static double calc_pow2(double n)
{
	return n*n;
}

//=============================================================================
// 
//=============================================================================
static double calc_distance(EFM *efm)
{
	return sqrt(calc_pow2((double)(efm->ex - efm->sx)) + calc_pow2((double)(efm->ey - efm->sy)));
}

//=============================================================================
// 
//=============================================================================
static void edit_message(EFM *efm)
{
	double scale;
	const wchar_t *unit;

	scale = upcall_getscale(NULL, &unit);
	sprintf(efm->msg, g_format, calc_distance(efm) * scale, unit);
}

///////////////////////////////////////////////////////////////////////////////
// プラグイン関数
//=============================================================================
// 必須
//=============================================================================
const char EXPORT *em_name(const char *lang)
{
	// メニュー項目・フィルターリスト用の名前文字列
	// lang は "jp"(日本語) か "en"(英語) が指定される

	// 名前を返す(例)
	return strcmp(lang, "jp") ? CAPTION_EN : CAPTION_JP;
}

//=============================================================================
// 必須
//=============================================================================
unsigned int EXPORT em_init(void(*redraw_pixel)(void), void(*redraw_view)(void), void(*close_ui)(void* instance), void(*set_videoframe)(size_t frame), void(*msg_func)(void), double(*get_scale)(int *axis_y, const wchar_t **unit))
{
	// アップコールを記録
	upcall_redraw_view = redraw_view;
	upcall_close_ui = close_ui;
	upcall_getscale = get_scale;

	// 共有ライブラリーロード時の初期化(必要なら)

	// 対応しているフォーマットと機能を返す(例)
	return TYPE_24|TYPE_UI;
}

//=============================================================================
// 必須
//=============================================================================
void EXPORT *em_create(void* instance)
{
	EFM *efm;

	// 領域確保
	efm = (EFM*)calloc(1, sizeof(EFM));

	// 内容の初期化
	if (efm) {
		efm->instance = instance;
	}

	// 終了
	return efm;
}

//=============================================================================
// 必須
//=============================================================================
void EXPORT em_delete(void *efmh)
{
	EFM *efm;

	// 内容の後始末
	efm = (EFM*)efmh;

	// 開放
	free(efm);
}

///////////////////////////////////////////////////////////////////////////////
// プラグイン関数(個別インスタンス毎の処理)
//=============================================================================
// 
//=============================================================================
#ifdef _WIN32
HWND EXPORT em_show(void *efmh, HWND parent)
#else
Window EXPORT em_show(void *efmh, Display *display, Window parent, int(**func)(XEvent xev, void *efmh))
#endif
{
	// サイドパネルでUI非表示にされた時の処理
	EFM *efm;
	XWND xwnd;

	// 管理データ取得
	efm = (EFM*)efmh;

	// 開く
	if (!window_enable(efm->dialog)) {
#ifdef _WIN32
		xwnd.window = parent;
#else
		xwnd.display = display;
		xwnd.window = parent;
		*func = event_func;
#endif
		edit_message(efm);
		efm->dialog = window_open(xwnd, g_caption, efm);
	}

	// 終了
	return efm->dialog.window;
}

//=============================================================================
// 
//=============================================================================
void EXPORT em_hide(void *efmh)
{
	// サイドパネルでUIを非表示にされた時の処理
	EFM *efm;

	// 管理データ取得
	efm = (EFM*)efmh;

	// 閉じる
	if (window_enable(efm->dialog)) {
		window_close(efm->dialog);
		memset(&efm->dialog, 0, sizeof(efm->dialog));
	}
}

//=============================================================================
// 
//=============================================================================
void EXPORT on_lang(void *efmh, const char *lang)
{
	// UIの表示言語を変えた時に通知される
	// lang は "jp"(日本語) か "en"(英語) が指定される
	EFM *efm;

	// 管理データ取得
	efm = (EFM*)efmh;

	// 言語変更(グローバル)
	if (!strcmp(lang, "jp")) {
		g_caption = CAPTION_JP;
		g_format = FORMAT_JP;
	}
	else {
		g_caption = CAPTION_EN;
		g_format = FORMAT_EN;
	}

	// 状態チェック
	if (!window_enable(efm->dialog)) {
		return;
	}

	// 言語変更
	window_caption(efm->dialog, g_caption);

	// テキスト編集
	edit_message(efm);

	// テキスト描画
	if (efm->dialog.window) {
		window_paint(efm->dialog);
	}
}

//=============================================================================
// 
//=============================================================================
int EXPORT on_filter(void *efmh, void *img, size_t w, size_t h, unsigned short bpp, size_t frame, size_t sx, size_t sy, size_t ex, size_t ey, int connect, int bg)
{
	// 画像の加工や計測等の実処理部分
	// img, w, h, bpp は画像バッファ情報
	// frame 0～ 動画・連番画像の場合は 0,1,…とフレーム番号が入る
	// sx, sy, ex, ey は処理領域
	// connectは 4or8 近傍
	// bgは2値画像処理時の背景色 0.黒 !0.白
	EFM *efm;

	// 管理データ取得
	efm = (EFM*)efmh;

	// スケール変更時に反映させるため、このイベントを処理する

	// テキスト編集
	edit_message(efm);

	// テキスト描画
	if (efm->dialog.window) {
		window_paint(efm->dialog);
	}

	// 画像を変更してるならON、変更していないならOFFを返す(例)
	return OFF;
}

//=============================================================================
// 
//=============================================================================
void EXPORT on_overlay_view(void *efmh, void(*pixel2screen)(double *x, double *y), double r, double g, double b)
{
	// 付加情報描画(必要なら)
	// 付加情報描画(必要なら)
	// pixel2screenはピクセル座標系からopen glの座標系に変換するためのツールと情報
	// r, g, b はガイド表示色

	// 描画はOpenGLを使う
	EFM *efm;
	double sx, sy, ex, ey;

	// 管理データ取得
	efm = (EFM*)efmh;

	// 座標系変換
	sx = (double)efm->sx;
	sy = (double)efm->sy;
	ex = (double)efm->ex;
	ey = (double)efm->ey;
	pixel2screen(&sx, &sy);
	pixel2screen(&ex, &ey);

	// 描画
	glColor3d(r, g, b);
	glBegin(GL_LINE_STRIP);
	glVertex2d(sx, sy);
	glVertex2d(ex, ey);
	glEnd();
}

//=============================================================================
// 
//=============================================================================
void EXPORT on_mouse_move(void *efmh, size_t x, size_t y, unsigned char opt_key)
{
	// フォーカス所有時に画像上をマウスが移動すると通知される
	// opt_keyは押されている特殊キーSHIFT(0x01)とCTRL(0x02)のOR値
	EFM *efm;

	// 管理データ取得
	efm = (EFM*)efmh;

	// マウスモードONの時のみ処理
	if (!efm->mouse_on) {
		return;
	}

	// 座標更新
	efm->ex = (int)x;
	efm->ey = (int)y;

	// テキスト編集
	edit_message(efm);

	// テキスト描画
	if (efm->dialog.window) {
		window_paint(efm->dialog);
	}

	// 画像描画
	upcall_redraw_view();
}

//=============================================================================
// 
//=============================================================================
void EXPORT on_mouse_lbutton_down(void *efmh, size_t x, size_t y, unsigned char opt_key)
{
	// フォーカス所有時に画像上をマウスの左ボタンが押されると通知される
	// opt_keyは押されている特殊キーSHIFT(0x01)とCTRL(0x02)のOR値
	EFM *efm;

	// 管理データ取得
	efm = (EFM*)efmh;

	// マウスモードON
	efm->mouse_on = ON;

	// 座標更新
	efm->sx = efm->ex = (int)x;
	efm->sy = efm->ey = (int)y;

	// テキスト編集
	edit_message(efm);

	// テキスト描画
	if (efm->dialog.window) {
		window_paint(efm->dialog);
	}

	// 画像描画
	upcall_redraw_view();
}

//=============================================================================
// 
//=============================================================================
void EXPORT on_mouse_lbutton_up(void *efmh, size_t x, size_t y, unsigned char opt_key)
{
	// フォーカス所有時に画像上をマウスの左ボタンが押されると通知される
	// opt_keyは押されている特殊キーSHIFT(0x01)とCTRL(0x02)のOR値
	EFM *efm;

	// 管理データ取得
	efm = (EFM*)efmh;

	// マウスモードOFF
	efm->mouse_on = OFF;
}
