﻿// Windows (VisualStudio)
//  cl /LD /Fetemplate.wfm template.c
//
// Linux (GCC)
//  gcc -O -shared -fPIC -o template.lfm template.c

///////////////////////////////////////////////////////////////////////////////
// ヘッダーファイル
#include <wchar.h>
#include <stdlib.h>
#include <string.h>
#ifdef _WIN32
#include <Windows.h>
#else
#include <X11/Xlib.h>
#include <GL/glx.h>
#endif
#include <GL/gl.h>

///////////////////////////////////////////////////////////////////////////////
// Windows用エクスポート宣言
#ifdef _WIN32
#ifndef _WIN64
#error 64bitバイナリーとしてビルドしてください
#endif
#define EXPORT __declspec(dllexport)
#else
#define EXPORT
#endif
///////////////////////////////////////////////////////////////////////////////
// 定数(タイプ)
#define	TYPE_8      0x0001
#define	TYPE_16     0x0002
#define	TYPE_32     0x0004
#define	TYPE_24     0x0008
#define	TYPE_48     0x0010
#define	TYPE_96     0x0020
#define	TYPE_192    0x0040
#define	TYPE_UI     0x0100

///////////////////////////////////////////////////////////////////////////////
// 定数
#define	OK          1
#define	NG          0
#define	ON          1
#define	OFF         0

///////////////////////////////////////////////////////////////////////////////
// 構造体
typedef struct _EFM {
	// インスタンス毎に保持するデータ
	void *instance;
} EFM;

///////////////////////////////////////////////////////////////////////////////
// 外部関数
static void   (*upcall_redraw_pixel)(void) = NULL;                          // UI操作で画像加工方法もしくはoverla_pixelの内容が変わった時に再描画要求する
static void   (*upcall_redraw_view)(void) = NULL;                           // UI操作でoverlay_viewの変わった時に再描画要求する
static void   (*upcall_close_ui)(void* instance) = NULL;                    // UIのクローズボタンでUIを閉じた事を通知する
static void   (*upcall_set_videoframe)(size_t frame) = NULL;                // 動画のフレームを変更する
static void   (*upcall_msgfunc)(void) = NULL;                               // GUIメッセージを1回処理する
static double (*upcall_getscale)(int* axis_y, const wchar_t **unit) = NULL; // 計測スケールと単位・座標系を取得 戻り値:ピクセル→実座標の係数 axis_y:1=Y軸下向き -1:Y軸上向き unit:単位文字列

///////////////////////////////////////////////////////////////////////////////
// 内部関数

///////////////////////////////////////////////////////////////////////////////
// プラグイン関数
//=============================================================================
// 必須
//=============================================================================
const char EXPORT *em_name(const char *lang)
{
	// メニュー項目・フィルターリスト用の名前文字列
	// lang は "jp"(日本語) か "en"(英語) が指定される

	// 名前を返す(例)
	return strcmp(lang, "jp") ? "Template" : "雛形";
}

//=============================================================================
// 必須
//=============================================================================
unsigned int EXPORT em_init(void (*redraw_pixel)(void), void (*redraw_view)(void), void (*close_ui)(void* instance), void (*set_videoframe)(size_t frame), void (*msg_func)(void), double (*get_scale)(int *axis_y, const wchar_t **unit))
{
	// アップコールを記録
	upcall_redraw_pixel = redraw_pixel;
	upcall_redraw_view = redraw_view;
	upcall_close_ui = close_ui;
	upcall_set_videoframe = set_videoframe;
	upcall_msgfunc = msg_func;
	upcall_getscale = get_scale;

	// 共有ライブラリーロード時の初期化(必要なら)

	// 対応しているフォーマットと機能を返す(例)
	return TYPE_24/*|TYPE_UI*/;
}

//=============================================================================
// 
//=============================================================================
void EXPORT em_exit(void)
{
	// 共有ライブラリー開放時の後始末(必要なら)
}

//=============================================================================
// 必須
//=============================================================================
void EXPORT *em_create(void* instance)
{
	EFM *efm;

	// 領域確保
	efm = (EFM*)calloc(1, sizeof(EFM));

	// 内容の初期化
	if (efm) {
		efm->instance = instance;
	}

	// 終了
	return efm;
}

//=============================================================================
// 必須
//=============================================================================
void EXPORT em_delete(void *efmh)
{
	// 内容の後始末
	EFM *efm;

	// 管理データ取得
	efm = (EFM*)efmh;

	// 開放
	free(efm);
}

///////////////////////////////////////////////////////////////////////////////
// プラグイン関数(個別インスタンス毎の処理)
//=============================================================================
// (ダイアログ使用時)
//=============================================================================
#ifdef _WIN32
HWND EXPORT em_show(void *efmh, HWND parent)
#else
Window EXPORT em_show(void *efmh, Display *display, Window parent, void (**func)(XEvent xev, void *efmh))
#endif
{
	// サイドパネルでUI非表示にされた時の処理
	EFM *efm;

	// 管理データ取得
	efm = (EFM*)efmh;

	// 終了
	return 0;
}

//=============================================================================
// (ダイアログ使用時)
//=============================================================================
void EXPORT em_hide(void *efmh)
{
	// サイドパネルでUIを非表示にされた時の処理
	EFM *efm;

	// 管理データ取得
	efm = (EFM*)efmh;
}

//=============================================================================
// (ダイアログ使用時)
//=============================================================================
void EXPORT em_uilock(void *efmh, int lock)
{
	// 一時的にUIを無効化/解除した時の処理
	EFM *efm;

	// 管理データ取得
	efm = (EFM*)efmh;
}

//=============================================================================
// 
//=============================================================================
void EXPORT on_enable(void *efmh, int enable)
{
	// サイドパネルでフィルターの有効/無効が切り替えられた時の処理(必要なら)
	EFM *efm;

	// 管理データ取得
	efm = (EFM*)efmh;
}

//=============================================================================
// 
//=============================================================================
void EXPORT on_lang(void *efmh, const char *lang)
{
	// UIの表示言語を変えた時に通知される
	// lang は "jp"(日本語) か "en"(英語) が指定される
	EFM *efm;

	// 管理データ取得
	efm = (EFM*)efmh;
}

//=============================================================================
// 
//=============================================================================
void EXPORT on_size(void *efmh, size_t w, size_t h, unsigned short bpp, size_t frames)
{
	// 処理する画像が変更された時に基本情報が通知される
	EFM *efm;

	// 管理データ取得
	efm = (EFM*)efmh;
}

//=============================================================================
// 
//=============================================================================
int EXPORT on_filter(void *efmh, void *img, size_t w, size_t h, unsigned short bpp, size_t frame, size_t sx, size_t sy, size_t ex, size_t ey, int connect, int bg)
{
	// 画像の加工や計測等の実処理部分
	// img, w, h, bpp は画像バッファ情報
	// frame 0～ 動画・連番画像の場合は 0,1,…とフレーム番号が入る
	// sx, sy, ex, ey は処理領域
	// connectは 4or8 近傍
	// bgは2値画像処理時の背景色 0.黒 !0.白
	EFM *efm;

	// 管理データ取得
	efm = (EFM*)efmh;

	// 画像を変更してるならON、変更していないならOFFを返す(例)
	return OFF;
}

//=============================================================================
// 
//=============================================================================
void EXPORT on_overlay_view(void *efmh, void (*pixel2screen)(double *x, double *y), double r, double g, double b)
{
	// 付加情報描画(必要なら)
	// pixel2screenはピクセル座標系からopen glの座標系に変換するためのツールと情報
	// r, g, b はガイド表示色

	// 描画はOpenGLを使う
	EFM *efm;

	// 管理データ取得
	efm = (EFM*)efmh;
}

//=============================================================================
// 
//=============================================================================
void EXPORT on_mouse_move(void *efmh, size_t x, size_t y, unsigned char opt_key)
{
	// 描画はOpenGLを使う、そのために上の変換で算出した座標を座標系(左上0,0)から()に変換する
	// opt_keyは押されている特殊キーSHIFT(0x01)とCTRL(0x02)のOR値
	EFM *efm;

	// 管理データ取得
	efm = (EFM*)efmh;
}

//=============================================================================
// 
//=============================================================================
void EXPORT on_mouse_lbutton_down(void *efmh, size_t x, size_t y, unsigned char opt_key)
{
	// フォーカス所有時に画像上をマウスの左ボタンが押されると通知される
	// opt_keyは押されている特殊キーSHIFT(0x01)とCTRL(0x02)のOR値
	EFM *efm;

	// 管理データ取得
	efm = (EFM*)efmh;
}

//=============================================================================
// 
//=============================================================================
void EXPORT on_mouse_lbutton_up(void *efmh, size_t x, size_t y, unsigned char opt_key)
{
	// フォーカス所有時に画像上をマウスの左ボタンが離されると通知される
	// opt_keyは押されている特殊キーSHIFT(0x01)とCTRL(0x02)のOR値
	EFM *efm;

	// 管理データ取得
	efm = (EFM*)efmh;
}

//=============================================================================
// 
//=============================================================================
void EXPORT on_mouse_rbutton_down(void *efmh, size_t x, size_t y, unsigned char opt_key)
{
	// フォーカス所有時に画像上をマウスの右ボタンが押されると通知される
	// opt_keyは押されている特殊キーSHIFT(0x01)とCTRL(0x02)のOR値
	EFM *efm;

	// 管理データ取得
	efm = (EFM*)efmh;
}

//=============================================================================
// 
//=============================================================================
void EXPORT on_mouse_rbutton_up(void *efmh, size_t x, size_t y, unsigned char opt_key)
{
	// フォーカス所有時に画像上をマウスの右ボタンが離されると通知される
	// opt_keyは押されている特殊キーSHIFT(0x01)とCTRL(0x02)のOR値
	EFM *efm;

	// 管理データ取得
	efm = (EFM*)efmh;
}

