﻿// Windows (VisualStudio)
//  cl /LD /Feeei.wim eei.c
//
// Linux (GCC)
//  gcc -O -shared -fPIC -o eei.lim eei.c

///////////////////////////////////////////////////////////////////////////////
// ヘッダーファイル
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

///////////////////////////////////////////////////////////////////////////////
// Windows用エクスポート宣言
#ifdef _WIN32
#ifndef _WIN64
#error 64bitバイナリーとしてビルドしてください
#endif
#define EXPORT __declspec(dllexport)
#else
#define EXPORT
#endif

///////////////////////////////////////////////////////////////////////////////
// 定数(タイプ)
#define	TYPE_8      0x0001
#define	TYPE_16     0x0002
#define	TYPE_32     0x0004
#define	TYPE_24     0x0008
#define	TYPE_48     0x0010
#define	TYPE_96     0x0020
#define	TYPE_192    0x0040
#define	TYPE_IMAGE  0x0100
#define	TYPE_MULTI  0x0200
#define	TYPE_VIDEO  0x0400

///////////////////////////////////////////////////////////////////////////////
// 定数
#define	OK              1
#define	NG              0
#define	YES             1
#define	NO              0
#define	OPEN_WRITE      1
#define	EMAGIC          0x484c
#define	EMAGIC_R        0x4c48
#define HEADER_SIZE 24

///////////////////////////////////////////////////////////////////////////////
// 構造体
typedef struct _FILECTRL {
	// 画像情報
	unsigned char bpp;
	unsigned int w;
	unsigned int h;
	unsigned int page;
	unsigned int ms;

	// 作業情報
	FILE *fp;
	int open_mode;          // 0.read 1.write
	int byte_order_match;   // 0.一致しない 1.一致する
	unsigned int next_page;
} FILECTRL;

///////////////////////////////////////////////////////////////////////////////
// 内部関数
//=============================================================================
// 反転2byte
//=============================================================================
static unsigned short rev2(unsigned short n)
{
	return ((n & 0x00ff) << 8) + ((n & 0xff00) >> 8);
}

//=============================================================================
// 反転4byte
//=============================================================================
static unsigned int rev4(unsigned int n)
{
	n = ((n & 0x0000ffff) << 16) |  ((n & 0xffff0000) >> 16);
	n = ((n & 0x00ff00ff) << 8) |  ((n & 0xff00ff00) >> 8);
	return n;
}

//=============================================================================
// 反転8byte
//=============================================================================
static unsigned int rev8(double n)
{
	long long ll;
	ll = (long long)n;
	ll = ((ll & 0x00000000ffffffff) << 32) |  ((ll & 0xffffffff00000000) >> 32);
	ll = ((ll & 0x0000ffff0000ffff) << 16) |  ((ll & 0xffff0000ffff0000) >> 16);
	ll = ((ll & 0x00ff00ff00ff00ff) << 8) |  ((ll & 0xff00ff00ff00ff00) >> 8);
	return (double)ll;
}

//=============================================================================
// ヘッダーロード
//=============================================================================
static int read_header(FILECTRL *fc)
{
	char magic[4], dummy;
	unsigned short file_e;
	unsigned short host_e = EMAGIC;

	// 識別子処理
	if (!fread(magic, 4, 1, fc->fp)) {
		return NG;
	}
	if (memcmp(magic, "EEI", 4)) {
		return NG;
	}

	// エンディアン
	if (!fread(&file_e, 2, 1, fc->fp)) {
		return NG;
	}
	if ((file_e != EMAGIC) && (file_e != EMAGIC_R)) {
		return NG;
	}
	fc->byte_order_match = (file_e==host_e)?YES:NO;

	// 0, bpp, w, h, page
	if (!fread(&dummy, 1, 1, fc->fp)) {
		return NG;
	}
	if (!fread(&fc->bpp, 1, 1, fc->fp)) {
		return NG;
	}
	if (!fread(&fc->w, 4, 1, fc->fp)) {
		return NG;
	}
	if (!fread(&fc->h, 4, 1, fc->fp)) {
		return NG;
	}
	if (!fread(&fc->page, 4, 1, fc->fp)) {
		return NG;
	}
	if (!fread(&fc->ms, 4, 1, fc->fp)) {
		return NG;
	}
	if (fc->byte_order_match == NO) {
		fc->bpp = rev2(fc->bpp);
		fc->w = rev4(fc->w);
		fc->h = rev4(fc->h);
		fc->ms = rev4(fc->ms);
	}
	if (!fc->w || !fc->h || !fc->page) {
		return NG;
	}
	switch (fc->bpp) {
	case 8:
	case 16:
	case 32:
	case 24:
	case 48:
	case 96:
	case 192:
		break;

	default:
		return NG;
	}

	// 終了
	return OK;
}

//=============================================================================
// ヘッダーセーブ
//=============================================================================
static int write_header(FILECTRL *fc)
{
	unsigned short e = EMAGIC;
	char dummy = 0;

	// 識別子
	if (!fwrite("EEI", 4, 1, fc->fp)) {
		return NG;
	}

	// エンディアン
	if (!fwrite(&e, 2, 1, fc->fp)) {
		return NG;
	}

	// 0, bpp, w, h, page
	if (!fwrite(&dummy, 1, 1, fc->fp)) {
		return NG;
	}
	if (!fwrite(&fc->bpp, 1, 1, fc->fp)) {
		return NG;
	}
	if (!fwrite(&fc->w, 4, 1, fc->fp)) {
		return NG;
	}
	if (!fwrite(&fc->h, 4, 1, fc->fp)) {
		return NG;
	}
	if (!fwrite(&fc->page, 4, 1, fc->fp)) {
		return NG;
	}
	if (!fwrite(&fc->ms, 4, 1, fc->fp)) {
		return NG;
	}

	// 終了
	return OK;
}

///////////////////////////////////////////////////////////////////////////////
// プラグイン関数
//=============================================================================
// このプログラムで処理するフォーマットのメタ情報
//=============================================================================
void EXPORT em_info(const char ***ext, unsigned int *type)
{
	// ***ext … 拡張子リストを返すポインタ
	// *type  … ファイルの種類を返すポインタ 1.静止画 2.マルチページ静止画 3.動画 4.全て
	static const unsigned int my_type = TYPE_IMAGE|TYPE_MULTI|TYPE_VIDEO|TYPE_8|TYPE_16|TYPE_32|TYPE_24|TYPE_48|TYPE_96|TYPE_192;
	static const char *ext_list[] = {"eei", NULL};

// 省略
//	// 引数チェック
//	if (!ext || !type) {
//		return;
//	}

	// メタ情報を返す
	*ext = ext_list;
	*type = my_type;
}

//=============================================================================
// ファイルを開く(ロード)
//=============================================================================
void EXPORT *em_open(const char *path, unsigned short *bpp, size_t *w, size_t *h, size_t *page, size_t *ms)
{
	// *path … ロードファイルパス
	// *bpp  … bit per pixelを返すポインタ [8, 16, 32, 24, 48, 96, 192]
	// *page … 画像枚数を返すポインタ
	// *w    … 水平方向ピクセル数を返すポインタ
	// *h    … 垂直方向ピクセル数を返すポインタ
	// *page … 画像枚数を返すポインタ
	// *ms   … 動画のフレーム間隔を返すポインタ
	FILECTRL *fc;

// 省略
//	// 引数チェック
//	if (!path || !*path) {
//		return NULL;
//	}
//	if (!bpp || !w || !h) {
//		return NULL;
//	}

	// 管理領域確保
	fc = (FILECTRL*)calloc(1, sizeof(FILECTRL));
	if (!fc) {
		return NULL;
	}

	// オープン
	fc->fp = fopen(path, "rb");
	if (!fc->fp) {
		free(fc);
		return NULL;
	}

	// ヘッダーロード
	if (!read_header(fc)) {
		free(fc);
		return NULL;
	}

	// 情報を返す
	*bpp = (unsigned short)fc->bpp;
	*w = (size_t)fc->w;
	*h = (size_t)fc->h;
	*page = (size_t)fc->page;
	*ms = (size_t)fc->ms;

	// 正常終了(管理領域を返す)
	return fc;
}

//=============================================================================
// 画像を読み込む
//=============================================================================
int EXPORT em_read(void *fh, void *buff, size_t page)
{
	// *fh   … ファイルハンドル
	// *buff … 画像データ
	// page  … 0～(静止画の場合は0固定)
	FILECTRL *fc;
	unsigned short *p2, n2;
	unsigned int *p4, n4;
	double *p8, n8;
	size_t i, cnt;
	long long size, pos;

// 省略
//	// 引数チェック
//	if (!fh) {
//		return NG;
//	}
//	if (!buff) {
//		return NG;
//	}

	// 引数取得
	fc = (FILECTRL*)fh;

	// 制限チェック
	if (page >= (size_t)fc->page) {
		return NG;
	}

	// シーク
	if (page != fc->next_page) {
		size = (fc->w * fc->h * (fc->bpp / 8)) * (long long)page;
		pos = HEADER_SIZE + size;	// オーバーフロー未チェック
#ifdef _WIN32
		if (_fseeki64(fc->fp, (__int64)pos, SEEK_SET)) {
#else
		if (fseek(fc->fp, (long)pos, SEEK_SET)) {
#endif
			return NG;
		}
		fc->next_page = page;
	}

	// 読み出し
	if ((fc->byte_order_match==YES) || (fc->bpp==8) || (fc->bpp=24)) {
		if (!fread(buff, fc->w*fc->h*(fc->bpp/8), 1, fc->fp)) {
			return NG;
		}
	}
	else if ((fc->bpp==16) | (fc->bpp==48)) {
		cnt = (size_t)(fc->w) * (size_t)(fc->h) * (size_t)(fc->bpp/16);
		p2 = (unsigned short*)buff;
		for (i=0; i<cnt; i++) {
			n2 = rev2(*p2++);
			if (!fread(&n2, 2, 1, fc->fp)) {
				return NG;
			}
		}
	}
	else if ((fc->bpp==32) | (fc->bpp==96)) {
		cnt = (size_t)(fc->w) * (size_t)(fc->h) * (size_t)(fc->bpp/32);
		p4 = (unsigned int*)buff;
		for (i=0; i<cnt; i++) {
			n4 = rev4(*p4++);
			if (!fread(&n4, 4, 1, fc->fp)) {
				return NG;
			}
		}
	}
	else if (fc->bpp==192) {
		cnt = (size_t)(fc->w) * (size_t)(fc->h) * (size_t)(fc->bpp/64);
		p8 = (double*)buff;
		for (i=0; i<cnt; i++) {
			n8 = rev8(*p8++);
			if (!fread(&n8, 8, 1, fc->fp)) {
				return NG;
			}
		}
	}
	else {
		return NG;
	}
	fc->next_page++;

	// 正常終了
	return OK;
}

//=============================================================================
// ファイルを作る(セーブ)
//=============================================================================
void EXPORT *em_create(const char *path, unsigned short bpp, size_t w, size_t h, size_t page, size_t ms)
{
	// *path … セーブファイルパス
	// bpp   … bit per pixel [8, 16, 32, 24, 48, 96, 192]
	// w     … 水平方向ピクセル数
	// h     … 垂直方向ピクセル数
	// page  … 画像枚数(静止画は1、マルチページや動画で最終枚数不明の場合は0)
	// ms    … 動画のフレーム間隔(静止画やマルチページの場合は0)
	FILECTRL *fc;

// 省略
//	// 引数チェック
//	if (!path || !*path) {
//		return NULL;
//	}
//	if (!bpp || !w || !h) {
//		return NULL;
//	}

	// 制限チェック
	if (w > (size_t)~(unsigned int)0) {
		return NG;
	}
	if (h > (size_t)~(unsigned int)0) {
		return NG;
	}
	if (ms > (size_t)~(unsigned int)0) {
		return NG;
	}
	if (bpp > (unsigned short)~(unsigned char)0) {
		return NG;
	}
	if ((bpp < 8) || (bpp % 8)) {
		return NG;
	}

	// 管理領域確保
	fc = (FILECTRL*)calloc(1, sizeof(FILECTRL));
	if (!fc) {
		return NULL;
	}

	// 管理情報初期化
	fc->w = w;
	fc->h = h;
	fc->bpp = bpp;
//closeで処理する	fc->page = page;
	fc->ms = ms;
	fc->open_mode = OPEN_WRITE;

	// オープン
	fc->fp = fopen(path, "wb");
	if (!fc->fp) {
		free(fc);
		return NULL;
	}

	// ヘッダーセーブ
	if (!write_header(fc)) {
		fclose(fc->fp);
		free(fc);
		return NULL;
	}

	// 正常終了(管理領域を返す)
	return fc;
}

//=============================================================================
// セーブ
//=============================================================================
int EXPORT em_write(void *fh, const void *buff)
{
	// *fh   … ファイルハンドル
	// *buff … 画像データ
	FILECTRL *fc;

// 省略
//	// 引数チェック
//	if (!fh) {
//		return NG;
//	}
//	if (!buff) {
//		return NG;
//	}

	// 引数取得
	fc = (FILECTRL*)fh;

	// 制限チェック
	if ((fc->page+1) == 0) {
		return NG;
	}

	// 書き込み
	if (!fwrite(buff, fc->w*fc->h*(fc->bpp/8), 1, fc->fp)) {
		return NG;
	}
	fc->page++;

	// 正常終了
	return OK;
}

//=============================================================================
// ファイルを閉じる
//=============================================================================
int EXPORT em_close(void *fh)
{
	// *fh … ファイルハンドル
	FILECTRL *fc;

// 省略
//	// 引数チェック
//	if (!fh) {
//		return NG;
//	}

	// 引数取得
	fc = (FILECTRL*)fh;

	// ページ数の書き込み
	if (fc->open_mode == OPEN_WRITE) {
#ifdef _WIN32
		_fseeki64(fc->fp, 16, SEEK_SET);
#else
		fseek(fc->fp, 16, SEEK_SET);
#endif
		fwrite(&fc->page, 4, 1, fc->fp);
	}

	// クローズ
	fclose(fc->fp);

	// 管理領域開放
	free(fh);

	// 正常終了
	return OK;
}
